@@ -1,10 +1,10 @@
 from ctREFPROP.ctREFPROP import REFPROPFunctionLibrary
-import REFPROPConnector.constants as constants
+import REFPROPConnector.Support.constants as constants
+from .Support.resources.file_handler import RP_EXEC
 from abc import ABC, abstractmethod
 from sty import fg, bg, ef, rs
 import warnings, sty
 
-
 GLOBALCounter = 0
 
 
@@ -29,8 +29,8 @@ class RefPropHandler:
         #   Molar Mass:         [kg/kmol]
         #   Heat Capacity:      [kJ/(kg*K)]
 
-        self.refprop = REFPROPFunctionLibrary(constants.RP_EXEC)
-        self.refprop.SETPATHdll(constants.RP_EXEC)
+        self.refprop = REFPROPFunctionLibrary(RP_EXEC)
+        self.refprop.SETPATHdll(RP_EXEC)
 
         self.fluids = fluids
         self.composition = composition
@@ -39,69 +39,46 @@ class RefPropHandler:
         self.TC = self.calculate("", "TC", 0, 0)
         self.PC = self.calculate("", "PC", 0, 0)
 
-    def __initialize_reference_state(self):
+    def set_reference_state(self, T_0=20, P_0=101325, T_0unit="C", P_0unit="Pa", old_unit_system=None):
 
         T_unit = self.return_units("T")
         P_unit = self.return_units("P")
 
-        T_0 = 20        # °C
-        P_0 = 101325    # Pa
+        if old_unit_system is not None:
 
-        if T_unit == "C":
+            T_unit_old = self.return_units("T", unit_system=old_unit_system)
+            P_unit_old = self.return_units("P", unit_system=old_unit_system)
 
-            # Celsius
-            self.T_0 = T_0
-
-        elif T_unit == "K":
-
-            # Kelvin
-            self.T_0 = T_0 + 273.15
+            self.T_0 = self.__convert_T(self.T_0, T_unit_old, T_unit)
+            self.P_0 = self.__convert_P(self.P_0, P_unit_old, P_unit)
 
         else:
 
-            # Fahrenheit
-            self.T_0 = T_0 * 9 / 5 + 32
-
-        if P_unit == "Pa":
-
-            # Pascal
-            self.P_0 = P_0
-
-        elif P_unit == "kPa":
-
-            # kiloPascal
-            self.P_0 = P_0 / 1000
-
-        elif P_unit == "bar":
-
-            # bar
-            self.P_0 = P_0 / 100000
-
-        elif P_unit == "MPa":
-
-            # bar
-            self.P_0 = P_0 / 100000
-
-        else:
-
-            # PSI
-            self.P_0 = P_0 / 6894.7572931783
+            self.T_0 = self.__convert_T(T_0, T_0unit, T_unit)
+            self.P_0 = self.__convert_P(P_0, P_0unit, P_unit)
 
         self.H_0 = self.calculate("TP", "H", self.T_0, self.P_0)
         self.S_0 = self.calculate("TP", "s", self.T_0, self.P_0)
 
     def calculate(self, str_in: str, str_out: str, a: float, b: float):
 
-        try:
-            self.refprop.SETFLUIDSdll('*'.join(self.fluids))
-            return self.refprop.REFPROP1dll(str_in, str_out, self.SI, 1, a, b, self.composition).c
-        except:
+        self.refprop.SETFLUIDSdll('*'.join(self.fluids))
+        return self.refprop.REFPROP1dll(str_in, str_out, self.SI, 1, a, b, self.composition).c
 
-            a= 20
+    def get_composition(self, phase, T, P):
 
-    def return_units(self, property_name):
+        # TODO
+        return self.composition
 
-        return_value = constants.get_units(property_name, self.unit_system)
+    def return_units(self, property_name, unit_system=None):
+
+        if unit_system is None:
+
+            return_value = constants.get_units(property_name, self.unit_system)
+
+        else:
+
+            return_value = constants.get_units(property_name, unit_system)
 
         if "Unknown" in return_value:
 
@@ -114,11 +91,15 @@ class RefPropHandler:
     @property
     def unit_system(self):
 
-        return self.__unit_system
+        try:
+            return self.__unit_system
+        except:
+            return None
 
     @unit_system.setter
     def unit_system(self, unit_system_input):
 
+        old_unit_system = self.unit_system
         self.__unit_system = unit_system_input
 
         try:
@@ -141,7 +122,106 @@ class RefPropHandler:
             warnings.warn(warning_message)
             self.__unit_system = "SI WITH C"
 
-        self.__initialize_reference_state()
+        self.set_reference_state(old_unit_system=old_unit_system)
+
+    @property
+    def T_0_in_K(self):
+
+        T_unit = self.return_units("T")
+        return self.__convert_T(self.T_0, T_unit, "K")
+
+    @staticmethod
+    def __convert_T_in_C(T, original_unit):
+
+        if original_unit == "C":
+
+            # Celsius
+            return T
+
+        elif original_unit == "K":
+
+            # Kelvin
+            return T - 273.15
+
+        else:
+
+            # Fahrenheit
+            return (T - 32) / 9 * 5
+
+    @staticmethod
+    def __convert_P_in_Pa(P, original_unit):
+
+        if original_unit == "Pa":
+
+            # Pascal
+            return P
+
+        elif original_unit == "kPa":
+
+            # kiloPascal
+            return P * 10 ** 3
+
+        elif original_unit == "bar":
+
+            # bar
+            return P * 10 ** 5
+
+        elif original_unit == "MPa":
+
+            # bar
+            return P * 10 ** 6
+
+        else:
+
+            # PSI
+            return P * 6894.7572931783
+
+    def __convert_T(self, T, original_unit, final_unit):
+
+        T_c = self.__convert_T_in_C(T, original_unit)
+        if final_unit == "C":
+
+            # Celsius
+            return T_c
+
+        elif final_unit == "K":
+
+            # Kelvin
+            return T_c + 273.15
+
+        else:
+
+            # Fahrenheit
+            return T_c * 9 / 5 + 32
+
+    def __convert_P(self, P, original_unit, final_unit):
+
+        P_pa = self.__convert_P_in_Pa(P, original_unit)
+
+        if final_unit == "Pa":
+
+            # Pascal
+            return P_pa
+
+        elif final_unit == "kPa":
+
+            # kiloPascal
+            return P_pa / 10 ** 3
+
+        elif final_unit == "bar":
+
+            # bar
+            return P_pa / 10 ** 5
+
+        elif final_unit == "MPa":
+
+            # bar
+            return P_pa / 10 ** 6
+
+        else:
+
+            # PSI
+            return P_pa / 6894.7572931783
 
 
 class ThermodynamicVariable:
@@ -180,7 +260,8 @@ class ThermodynamicVariable:
 class AbstractThermodynamicPoint(ABC):
 
     @classmethod
-    def init_from_fluid(cls, fluids: list, composition: list, other_variables="all", calculate_on_need="all", unit_system="SI WITH C"):
+    def init_from_fluid(cls, fluids: list, composition: list, other_variables="all", calculate_on_need="all",
+                        unit_system="SI WITH C"):
 
         RP = RefPropHandler(fluids, composition, unit_system)
         return cls(RP, other_variables, calculate_on_need)
@@ -188,10 +269,16 @@ class AbstractThermodynamicPoint(ABC):
     def __init__(self, refprop: RefPropHandler, other_variables="all", calculate_on_need="all"):
 
         self.RPHandler = refprop
+        self.inputs = {
+
+            "other_variables": other_variables,
+            "calculate_on_need": calculate_on_need
+
+        }
 
         self.__initialize_state_variables()
         self.__initialize_other_variables(other_variables)
-        self.reset_state_var_order()
+        self.__reset_state_var_order()
 
         self.calculated_variables = list()
         self.__initialize_calculate_on_need_variables(calculate_on_need)
@@ -278,6 +365,13 @@ class AbstractThermodynamicPoint(ABC):
 
     def __calculate_variable(self, variable):
 
+        if self.calculation_ready:
+
+            variable.value = self.__calculate_direct(variable.refprop_name)
+            self.calculated_variables.append(variable)
+
+    def __calculate_direct(self, REFPROP_CODE):
+
         global GLOBALCounter
 
         if self.calculation_ready:
@@ -287,12 +381,18 @@ class AbstractThermodynamicPoint(ABC):
             not_none_variables = self.not_none_variables
             input_str = not_none_variables[0].refprop_name + not_none_variables[1].refprop_name
 
-            variable.value = self.RPHandler.calculate(input_str,
-                                                      variable.refprop_name,
-                                                      not_none_variables[0].value,
-                                                      not_none_variables[1].value)
+            value = self.RPHandler.calculate(
 
-            self.calculated_variables.append(variable)
+                input_str,
+                REFPROP_CODE,
+                not_none_variables[0].value,
+                not_none_variables[1].value
+
+            )
+
+            return value
+
+        return None
 
     @abstractmethod
     def other_calculation(self):
@@ -305,14 +405,17 @@ class AbstractThermodynamicPoint(ABC):
         for variable in self.state_var_list:
 
             if variable.refprop_name == input_refprop_name:
+
                 variable.value = variable_value
                 variable.order = 0
+
                 break
 
         self.state_var_list.sort()
-        self.reset_state_var_order()
+        self.__reset_state_var_order()
 
         if self.calculation_ready:
+
             self.__calculate()
 
     def get_variable(self, variable_name: str):
@@ -322,16 +425,17 @@ class AbstractThermodynamicPoint(ABC):
         if variable is not None:
 
             if variable not in self.calculated_variables:
+
                 self.__calculate_variable(variable)
 
             return variable.value
 
         if variable_name == "exergy":
 
-            h = self.get_variable("h")
-            s = self.get_variable("s")
+            dh = self.get_variable("h") - self.RPHandler.H_0
+            ds = self.get_variable("s") - self.RPHandler.S_0
 
-            return (h - self.RPHandler.H_0) - self.RPHandler.T_0 * (s - self.RPHandler.S_0)
+            return dh - self.RPHandler.T_0_in_K * ds
 
         try:
 
@@ -341,10 +445,57 @@ class AbstractThermodynamicPoint(ABC):
 
             return None
 
+    def get_derivative(self, num_name: str, den_name: str, fix_name: str):
+
+        if self.calculation_ready:
+
+            info = constants.get_derivative_info(num_name, den_name, fix_name)
+
+            if info is not None:
+
+                rp_codes = info.refprop_codes
+
+                if info.is_fraction:
+
+                    a = self.__calculate_direct(rp_codes[0])
+                    b = self.__calculate_direct(rp_codes[1])
+
+                    return - a / b
+
+                elif info.is_inverse:
+
+                    return 1 / self.__calculate_direct(rp_codes[0])
+
+                else:
+
+                    return self.__calculate_direct(rp_codes[0])
+
+        return None
+
+    def get_composition(self, phase):
+
+        if self.calculation_ready:
+
+            Q = self.get_variable("Q")
+
+            if 0 < Q < 1:
+
+                T = self.get_variable("T")
+                P = self.get_variable("P")
+                self.RPHandler.get_composition(phase, T, P)
+
+            else:
+
+                return self.composition
+
     def get_unit(self, variable_name: str):
 
         return self.RPHandler.return_units(variable_name)
 
+    def set_unit_system(self, unit_system):
+
+        self.RPHandler.unit_system = unit_system
+
     def __get_variable_from_name(self, variable_name: str):
 
         input_refprop_name = constants.get_refprop_name(variable_name.lower())
@@ -356,7 +507,7 @@ class AbstractThermodynamicPoint(ABC):
 
         return None
 
-    def reset_state_var_order(self):
+    def __reset_state_var_order(self):
 
         counter = 1
         for variable in self.state_var_list:
@@ -367,6 +518,7 @@ class AbstractThermodynamicPoint(ABC):
     def calculation_ready(self):
 
         counter = 0
+
         for variable in self.state_var_list:
 
             if not variable.is_empty:
@@ -411,7 +563,7 @@ class AbstractThermodynamicPoint(ABC):
         return return_list
 
     @property
-    def state_variables_refprop_names(self)->list:
+    def state_variables_refprop_names(self) -> list:
 
         return_list = list()
         for variable in self.state_var_list:
@@ -434,6 +586,27 @@ class AbstractThermodynamicPoint(ABC):
 
         return self.RPHandler.composition
 
+    @property
+    def reference_state(self):
+
+        reference = self.duplicate()
+        reference.set_variable("T", self.RPHandler.T_0)
+        reference.set_variable("P", self.RPHandler.P_0)
+
+        return reference
+
+    @reference_state.setter
+    def reference_state(self, state):
+
+        self.RPHandler.set_reference_state(
+
+            T_0 = state.get_variable("T"),
+            P_0 = state.get_variable("P"),
+            T_0unit = state.get_unit("T"),
+            P_0unit = state.get_unit("P")
+
+        )
+
     @composition.setter
     def composition(self, new_composition: list):
 
@@ -448,6 +621,31 @@ class AbstractThermodynamicPoint(ABC):
         if reset:
             GLOBALCounter = 0
 
+    @abstractmethod
+    def duplicate(self):
+        """
+            used to create a copy of the point for further calculations
+            in programming it in subclasses use the input dictionary to recover original inputs
+
+            don't forget to copy the current state defined by the user:
+
+                self.copy_state_to(new_point)
+
+        """
+        pass
+
+    def copy_state_to(self, target_point):
+
+        for variable in self.state_var_list[:2]:
+
+            if variable.is_empty:
+                break
+
+            else:
+                target_point.set_variable(variable.refprop_name, variable.value)
+
+        return target_point
+
     def list_properties(self):
 
         string_to_display = """
@@ -553,9 +751,9 @@ class AbstractThermodynamicPoint(ABC):
         variable_format = " {:20s}"
 
         name_bold = ef.bold + name_format + ef.rs
-        name_bold_back = bg.yellow + ef.bold + name_format + ef.rs + bg.rs
+        name_back = bg.li_yellow + name_format + bg.rs
         variable_bold = ef.bold + variable_format + ef.rs
-        variable_back = bg.yellow + variable_format + bg.rs
+        variable_back = bg.li_yellow + variable_format + bg.rs
 
         string_to_display = "\n" + name_format.format("") + " ".join(
 
@@ -567,10 +765,11 @@ class AbstractThermodynamicPoint(ABC):
 
             if unit_system == self.RPHandler.unit_system:
 
-                string_to_display += name_bold_back.format(unit_system)
+                string_to_display += name_back.format(unit_system)
                 string_to_display += "".join(
 
-                    variable_back.format(constants.get_units(variable.refprop_name, unit_system)) for variable in variable_list
+                    variable_back.format(constants.get_units(variable.refprop_name, unit_system)) for variable in
+                    variable_list
 
                 )
 
@@ -591,7 +790,8 @@ class AbstractThermodynamicPoint(ABC):
 
 class ThermodynamicPoint(AbstractThermodynamicPoint):
 
-    def __init__(self, fluids: list, composition: list, other_variables="all", calculate_on_need="all", unit_system="SI WITH C"):
+    def __init__(self, fluids: list, composition: list, other_variables="all", calculate_on_need="all",
+                 unit_system="SI WITH C"):
 
         RP = RefPropHandler(fluids, composition, unit_system)
         super().__init__(RP, other_variables=other_variables, calculate_on_need=calculate_on_need)
@@ -599,6 +799,22 @@ class ThermodynamicPoint(AbstractThermodynamicPoint):
     def other_calculation(self):
         pass
 
-    def init_from_fluid(cls, fluids: list, composition: list, other_variables="all", calculate_on_need="all", unit_system="SI WITH C"):
+    def init_from_fluid(cls, fluids: list, composition: list, other_variables="all", calculate_on_need="all",
+                        unit_system="SI WITH C"):
 
         return ThermodynamicPoint(fluids, composition, other_variables, calculate_on_need, unit_system)
+
+    def duplicate(self):
+
+        tp = ThermodynamicPoint(
+
+            self.RPHandler.fluids,
+            self.RPHandler.composition,
+            unit_system=self.RPHandler.unit_system,
+            other_variables=self.inputs["other_variables"],
+            calculate_on_need=self.inputs["calculate_on_need"]
+
+        )
+
+        self.copy_state_to(tp)
+        return tp
\ No newline at end of file
